/*
 * Copyright (C) 2005 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "gcc-compat.h"
#include "para.h"

#include "list.h"
#include "filter.h"
#include <mad.h>

#define MP3_OUTBUF_SIZE 100 * 1000
#define MAD_TO_SHORT(f) (f) >= MAD_F_ONE?  SHRT_MAX :\
(f) <= -MAD_F_ONE? -SHRT_MAX : (signed short) ((f) >> (MAD_F_FRACBITS - 15))

struct mp3dec_data {
	struct mad_stream stream;
	struct mad_frame frame;
	struct mad_synth synth;
	mad_timer_t timer;
};

static ssize_t mp3dec(char *inbuffer, size_t len, struct filter_node *fn,
		struct filter_chain_info *fci)
{
	int i, ret;
	struct mp3dec_data *md = fn->private_data;

	mad_stream_buffer(&md->stream, inbuffer, len);
	md->stream.error = 0;
next_frame:
	ret = mad_frame_decode(&md->frame, &md->stream);
	if (ret) {
		if (MAD_RECOVERABLE(md->stream.error) || md->stream.error == MAD_ERROR_BUFLEN)
			goto out;
		PARA_ERROR_LOG("fatal: ret = %d, loaded = %d\n", ret, fn->loaded);
		return -1;
	}
	mad_timer_add(&md->timer, md->frame.header.duration);
	mad_synth_frame(&md->synth, &md->frame);
	fci->samplerate = md->frame.header.samplerate;
	fci->channels = MAD_NCHANNELS(&md->frame.header);
//	PARA_DEBUG_LOG("%d channels, %d KHz\n", fci->channels, fci->samplerate);

	for (i = 0; i < md->synth.pcm.length; i++) {
		/* output format: unsigned 16 bit little endian */
		signed short s = MAD_TO_SHORT(md->synth.pcm.samples[0][i]);
		fn->buf[fn->loaded++] = s & 0xff;
		fn->buf[fn->loaded++] = s >> 8;
		if (MAD_NCHANNELS(&md->frame.header) == 2) { /* stereo */
			s = MAD_TO_SHORT(md->synth.pcm.samples[1][i]);
			fn->buf[fn->loaded++] = s & 0xff;
			fn->buf[fn->loaded++] = s >> 8;
		}
		if (fn->loaded != fn->bufsize) /* output buffer not full */
			continue;
		PARA_ERROR_LOG("output buffer full: %d\n", fn->loaded);
			return -1;
	}
	goto next_frame;
out:
	if (md->stream.next_frame) { /* we still have some data */
		size_t off = md->stream.bufend - md->stream.next_frame;
//		PARA_DEBUG_LOG("converted %d, %d input bytes, %d output bytes\n",
//			len - off, off, self->loaded);
		return len - off;
	}
	return len;
}

static void mp3dec_close(struct filter_node *fn)
{
	struct mp3dec_data *md = fn->private_data;

	mad_synth_finish(&md->synth);
	mad_frame_finish(&md->frame);
	mad_stream_finish(&md->stream);

	free(fn->buf);
	fn->buf = NULL;
	free(md);
	fn->private_data = NULL;
}

static int mp3dec_open(struct filter_node *fn)
{
	fn->private_data = (struct mp3dec_data*) para_malloc(sizeof(struct mp3dec_data));
	struct mp3dec_data *md = fn->private_data;

	mad_stream_init(&md->stream);
	mad_frame_init(&md->frame);
	mad_synth_init(&md->synth);
	mad_timer_reset(&md->timer);
	fn->loaded = 0;
	fn->bufsize = MP3_OUTBUF_SIZE;
	fn->buf = para_malloc(fn->bufsize);
	return 1;
}

int mp3dec_init(struct filter *f)
{
	f->open = mp3dec_open;
	f->convert = mp3dec;
	f->close = mp3dec_close;
	return 1;
}
